@@ -1,5 +1,5 @@
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
- * This is GNU GO, a Go program. Contact gnugo@gnu.org, or see       *
+ * This is GNU Go, a Go program. Contact gnugo@gnu.org, or see       *
  * http://www.gnu.org/software/gnugo/ for more information.          *
  *                                                                   *
  * Copyright 1999, 2000, 2001, 2002 by the Free Software Foundation. *
@@ -23,13 +23,10 @@
 #include "liberty.h"
 #include "patterns.h"
 
-#include <stdlib.h>
-#include <assert.h>
-
 
 #define TRYMOVE(pos, color) trymove(pos, color, "helper", NO_MOVE, EMPTY, NO_MOVE)
-#define OFFSET(x, y) offset(x, y, move, transformation)
-#define ARGS struct pattern *pattern, int transformation, int move, int color
+#define OFFSET_BY(x, y) AFFINE_TRANSFORM(OFFSET(x,y), trans, move)
+#define ARGS struct pattern *pattern, int trans, int move, int color
 
 
 /* This file contains helper functions which assist the pattern matcher.
@@ -60,9 +57,9 @@ basic_cut_helper(ARGS)
   int acolor, ccolor;
   UNUSED(pattern);
   
-  apos = OFFSET(0, -1);  /* O to west */
-  bpos = OFFSET(-1, 0);  /* O to north */
-  cpos = OFFSET(-1, -1); /* X to northwest */
+  apos = OFFSET_BY(0, -1);  /* O to west */
+  bpos = OFFSET_BY(-1, 0);  /* O to north */
+  cpos = OFFSET_BY(-1, -1); /* X to northwest */
 
   acolor = board[apos];
   ccolor = OTHER_COLOR(acolor);
@@ -108,14 +105,14 @@ jump_out_helper(ARGS)
 {
   int own_eyespace;
 
-  UNUSED(transformation); UNUSED(pattern);
+  UNUSED(trans); UNUSED(pattern);
 
   if (color == WHITE)
     own_eyespace = (white_eye[move].color == WHITE_BORDER);
   else
     own_eyespace = (black_eye[move].color == BLACK_BORDER);
   
-  if (influence_area_color(move) != color && !own_eyespace)
+  if (whose_area(OPPOSITE_INFLUENCE(color), move) != color && !own_eyespace)
     return 1;
   else
     return 0;
@@ -130,8 +127,8 @@ jump_out_helper(ARGS)
 int 
 jump_out_far_helper(ARGS)
 {
-  if (influence_area_color(move) != OTHER_COLOR(color))
-    return jump_out_helper(pattern, transformation, move, color);
+  if (whose_area(OPPOSITE_INFLUENCE(color), move) != OTHER_COLOR(color))
+    return jump_out_helper(pattern, trans, move, color);
   else
     return 0;
 }
@@ -143,7 +140,7 @@ jump_out_far_helper(ARGS)
 int
 high_handicap_helper(ARGS)
 {
-  UNUSED(transformation); UNUSED(pattern); UNUSED(move);
+  UNUSED(trans); UNUSED(pattern); UNUSED(move);
   
   return !doing_scoring && stones_on_board(OTHER_COLOR(color)) == 0;
 }
@@ -158,7 +155,7 @@ high_handicap_helper(ARGS)
 int
 reinforce_helper(ARGS)
 {
-  UNUSED(transformation); UNUSED(pattern);
+  UNUSED(trans); UNUSED(pattern);
   
   return (!doing_scoring
 	  && !lively_dragon_exists(OTHER_COLOR(color)) 
@@ -201,9 +198,9 @@ throw_in_atari_helper(ARGS)
   int libs[2];
   UNUSED(pattern);
   
-  apos = OFFSET(0, 1);
-  cpos = OFFSET(-1, 1);
-  dpos = OFFSET(1, 1);
+  apos = OFFSET_BY(0, 1);
+  cpos = OFFSET_BY(-1, 1);
+  dpos = OFFSET_BY(1, 1);
 
   /* Find second liberty of the stone a. */
   findlib(apos, 2, libs);
@@ -227,7 +224,8 @@ throw_in_atari_helper(ARGS)
   }
 
   /* The followup is to capture the "a" string. Estimate the value to
-     twice the size. */
+   * twice the size.
+   */
   add_followup_value(move, 2 * worm[apos].effective_size);
   TRACE("...followup value %f\n", 2 * worm[apos].effective_size);
 
@@ -235,35 +233,6 @@ throw_in_atari_helper(ARGS)
 }
 
 
-/*
- *
- * Prevent misreporting of a as lunch for b.
- * To be used in autohelper action line. E.g.
- *
- *  XO|          ba|
- *  O*|          O*|
- *  oo|          oo|
- *  ?o|          ?o|
- *  
- *  >not_lunch(a,b);
- */
-
-int
-not_lunch_helper(int apos, int bpos)
-{
-  if (worm[apos].size > 2)
-    return 0;
-
-  /* Tell the move generation code about the change in status. */
-  remove_lunch(bpos, apos);
-  
-  if (DRAGON2(bpos).lunch == apos)
-    DRAGON2(bpos).lunch = NO_MOVE;
-
-  return 0;
-}
-  
-
 /* This is intended for use in autohelpers. */
 
 /* Check whether the string at (str) can attack any surrounding
@@ -302,7 +271,7 @@ ugly_cutstone_helper(ARGS)
   UNUSED(pattern);
   UNUSED(color);
   
-  apos = OFFSET(-1, -1);
+  apos = OFFSET_BY(-1, -1);
   
   worm[apos].cutstone++;
   return 0;
@@ -328,11 +297,11 @@ cutstone2_helper(ARGS)
   UNUSED(pattern);
   UNUSED(color);
   
-  apos = OFFSET(-1, -1);
-  bpos = OFFSET(-1,  0);
-  cpos = OFFSET( 0, -1);
+  apos = OFFSET_BY(-1, -1);
+  bpos = OFFSET_BY(-1,  0);
+  cpos = OFFSET_BY( 0, -1);
 
-  if (worm[apos].defend_codes[0] == 0)
+  if (worm[apos].defense_codes[0] == 0)
     return 0;
   
   dpos = worm[apos].defense_points[0];
@@ -393,6 +362,19 @@ threaten_to_save_helper(int move, int str)
 }
 
 
+/* For use in autohelpers.
+ *
+ * Adds a reverse followup value if the opponent's move here would threaten
+ * to capture (str).
+ */
+void
+prevent_attack_threat_helper(int move, int str)
+{
+  add_reverse_followup_value(move, 2.0 * worm[str].effective_size);
+  TRACE("...reverse followup value %f\n", 2.0 * worm[str].effective_size);
+}
+
+
 /*
  * This is intended for use in autohelpers.
  *
@@ -427,7 +409,7 @@ threaten_to_capture_helper(int move, int str)
   
   adj = chainlinks2(str, adjs, 1);
   for (k = 0; k < adj; k++)
-    if (worm[adjs[k]].defend_codes[0] != 0
+    if (worm[adjs[k]].defense_codes[0] != 0
 	&& !does_defend(move, adjs[k]))
       return;
 
@@ -443,6 +425,25 @@ threaten_to_capture_helper(int move, int str)
     }
     popgo();
   }
+
+  /* In addition to the move found by find_defense(), also try all
+   * chain breaking moves in the same way.
+   */
+  adj = chainlinks2(str, adjs, 1);
+  for (k = 0; k < adj; k++) {
+    int lib;
+    findlib(adjs[k], 1, &lib);
+    if (TRYMOVE(lib, board[str])) {
+      if (!attack(str, NULL)
+	  && (board[move] == EMPTY || attack(move, NULL) != 0)) {
+	popgo();
+	popgo();
+	return;
+      }
+      popgo();
+    }
+  }
+
   popgo();
   
   add_followup_value(move, 2.0 * worm[str].effective_size);
@@ -469,7 +470,7 @@ defend_against_atari_helper(int move, int str)
   /* No value if the string can capture out of atari. */
   adj = chainlinks2(str, adjs, 1);
   for (k = 0; k < adj; k++)
-    if (worm[adjs[k]].defend_codes[0] != 0
+    if (worm[adjs[k]].defense_codes[0] != 0
 	&& !does_defend(move, adjs[k]))
       return;
 
@@ -593,13 +594,13 @@ backfill_helper(int apos, int bpos, int cpos)
 int
 owl_threatens_attack(int apos, int bpos)
 {
-  if (dragon[bpos].owl_status == CRITICAL
-      && dragon[bpos].owl_attack_point == apos)
+  if (DRAGON2(bpos).owl_status == CRITICAL
+      && DRAGON2(bpos).owl_attack_point == apos)
     return 1;
   
-  if (dragon[bpos].owl_threat_status == CAN_THREATEN_ATTACK)
-    if (dragon[bpos].owl_attack_point == apos
-	|| dragon[bpos].owl_second_attack_point == apos)
+  if (DRAGON2(bpos).owl_threat_status == CAN_THREATEN_ATTACK)
+    if (DRAGON2(bpos).owl_attack_point == apos
+	|| DRAGON2(bpos).owl_second_attack_point == apos)
       return 1;
   
   return 0;
@@ -724,31 +725,111 @@ connect_and_cut_helper2(int Apos, int bpos, int cpos, int color)
 
 
 
-/* replaces macro with assert-enabled function */
-int 
-dragon_weak(int pos)
+void
+test_attack_either_move(int move, int color, int worma, int wormb)
 {
-  ASSERT_ON_BOARD1(pos);
-  /* FIXME: This should not happen, but avoids a crash.  What is
-   *   the proper fix for calling this at stackp != 0 ?
-   */
-  if (dragon[pos].id < 0 || dragon[pos].id >= number_of_dragons)
-     return 1;
-  return DRAGON2(pos).safety != ALIVE
-         && DRAGON2(pos).safety != STRONGLY_ALIVE \
-	 && DRAGON2(pos).safety != INVINCIBLE;
+  ASSERT_ON_BOARD1(move);
+  ASSERT1(board[move] == EMPTY, move);
+  ASSERT1(board[worma] == OTHER_COLOR(color)
+          && board[wormb] == OTHER_COLOR(color), move);
+
+  if (!defend_both(worma, wormb)) {
+    if (0)
+      gprintf("%1m: Reject attack_either_move for %1m, %1m (can't defend both)\n",
+	      move, worma, wormb);
+    return;
+  }
+  if (trymove(move, color, "test_attack_either_move", worma, EMPTY, NO_MOVE)) {
+    if (board[worma] == OTHER_COLOR(color)
+	&& board[wormb] == OTHER_COLOR(color)) {
+      if (!defend_both(worma, wormb))
+        add_either_move(move, ATTACK_STRING, worma, ATTACK_STRING, wormb);
+      else {
+	if (0)
+	  gprintf("%1m: Rej. attack_either_move for %1m & %1m (doesn't work)\n",
+		  move, worma, wormb);
+      }
+    }
+    else
+      if (0)
+	gprintf("%1m: Rej. attack_either_move for %1m & %1m (captured directly)\n",
+		move, worma, wormb);
+    popgo();
+  }
 }
 
+/* True if str is adjacent to a stone in atari, which is tactically
+ * defendable.
+ */
+int
+adjacent_to_defendable_stone_in_atari(int str)
+{
+  int adj;
+  int adjs[MAXCHAIN];
+  int k;
+
+  adj = chainlinks2(str, adjs, 1);
+  for (k = 0; k < adj; k++)
+    if (attack_and_defend(adjs[k], NULL, NULL, NULL, NULL))
+      return 1;
+  
+  return 0;
+}
+
+
+void
+backfill_replace(int move, int str)
+{
+  int defense_move = NO_MOVE;
+
+  if (TRYMOVE(move, OTHER_COLOR(board[str]))) {
+    if (attack_and_defend(str, NULL, NULL, NULL, &defense_move)) {
+      /* Must undo the trymove before adding the replacement move. */
+      popgo();
+      add_replacement_move(move, defense_move);
+    }
+    else
+      popgo();
+  }
+}
+
+
+/* True if
+ * 1. White to move.
+ * 2. All white stones look dead.
+ * 3. Less than 40% of the board is filled or less than 20% of the
+ *    board is filled with white stones.
+ *
+ * This is intended for patterns forcing white to thrash around early
+ * in high handicap games, instead of passing because it looks like no
+ * stones can live.
+ */
+int 
+thrash_around_helper(ARGS)
+{
+  int d;
+  UNUSED(pattern);
+  UNUSED(trans);
+  UNUSED(move);
+  
+  if (doing_scoring
+      || (stones_on_board(BLACK | WHITE) > board_size * board_size * 2 / 5
+	  && stones_on_board(WHITE) > board_size * board_size / 5)
+      || color == BLACK)
+    return 0;
+
+  for (d = 0; d < number_of_dragons; d++)
+    if (DRAGON(d).color == WHITE
+	&& DRAGON(d).status != DEAD)
+      return 0;
+
+  return 1;
+}
+
+
 /*
  * LOCAL Variables:
  * tab-width: 8
  * c-basic-offset: 2
  * End:
  */
-
-
-void
-set_value_helper(struct pattern *patt, int value)
-{
-  patt->value = value;
-}
